from typing import Optional, Dict, Any, List, Union, ClassVar, Callable
import logging
import websockets
import pickle
import json
from dataclasses import dataclass, asdict
from inspect import isclass, signature, getdoc, Signature, Parameter, _empty
from concurrent.futures import ThreadPoolExecutor
from rpc_gateway import errors, gateway, messages_pb2, utils
from rpc_gateway.utils import await_sync

logger = logging.getLogger(__name__)


def get_attribute_path(obj, attribute_path: List[str]):
    if len(attribute_path) == 1:
        return getattr(obj, attribute_path[0])

    return get_attribute_path(getattr(obj, attribute_path[0]), attribute_path[1:])


def set_attribute_path(obj, attribute_path: List[str], value: Any):
    if len(attribute_path) == 1:
        return setattr(obj, attribute_path[0], value)

    return get_attribute_path(getattr(obj, attribute_path[0]), attribute_path[1:])


def is_json_serializable(value: Any) -> bool:
    try:
        json.dumps(value)
        return True
    except TypeError:
        return False


@dataclass
class Instance:
    name: str
    instance: Any
    group: str
    locked: bool


@dataclass
class Argument:
    name: str
    default: Any
    serializable: bool

    @classmethod
    def from_parameter(cls, parameter: Parameter) -> 'Argument':
        default = None if parameter.default is _empty else parameter.default
        serializable = is_json_serializable(default)

        return cls(
            name=parameter.name,
            default=default if serializable else str(default),
            serializable=serializable
        )


@dataclass
class Attribute:
    name: str
    value: Any
    serializable: bool

    @classmethod
    def from_value(cls, name: str, value: Any) -> 'Attribute':
        serializable = is_json_serializable(value)

        return cls(
            name=name,
            value=value if serializable else str(value),
            serializable=serializable
        )


@dataclass
class Method:
    name: str
    description: str
    arguments: List[Argument]

    @classmethod
    def from_method(cls, name: str, method: Callable) -> 'Method':
        method_signature = signature(method)
        return cls(
            name=name,
            description=getdoc(method),
            arguments=[Argument.from_parameter(param) for param in method_signature.parameters.values()]
        )


@dataclass
class InstanceMetadata:
    name: str
    locked: bool
    description: str
    attributes: List[Attribute]
    methods: List[Method]

    @classmethod
    def from_instance(cls, instance: Instance) -> 'InstanceMetadata':
        attributes = []
        methods = []

        for name in dir(instance.instance):
            if not name.startswith('_'):
                value = getattr(instance.instance, name)

                if callable(value):
                    methods.append(Method.from_method(name, value))
                else:
                    attributes.append(Attribute.from_value(name, value))

        return cls(
            name=instance.name,
            description=getdoc(instance.instance),
            attributes=attributes,
            methods=methods,
            locked=instance.locked
        )


class Server(gateway.GatewayClient):
    def __init__(self, gateway_url: str = 'ws://localhost:8888', max_workers: int = 5, instances: Optional[Dict[str, Instance]] = None, auth_key: str = 'DEFAULT_KEY'):
        super().__init__(gateway_url=gateway_url, auth_key=auth_key)
        self.instances = {} if instances is None else instances
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

    async def _on_start(self):
        await self._register_gateway_instances(self.instances)

    async def _on_stop(self):
        self.instances = {}

    async def _register_gateway_instances(self, instances: Dict[str, Instance]):
        instance_data = [(instance.name, instance.group) for name, instance in instances.items()]
        await self.websocket_connection.send_request(messages_pb2.Method.REGISTER, data=instance_data)

    def register(self, name: str, instance: Any, group: Optional[str] = None) -> Any:
        self.logger.info(f'Registering instance: {name}')

        if name is None or name == '':
            raise errors.InvalidInstanceError(f'Cannot register instance with name: {name}')

        if name in self.instances:
            raise errors.InstanceAlreadyRegisteredError(f'Instance already registered with name: {name}')

        if group is None:
            group = instance.__class__.__name__

        self.instances[name] = Instance(name=name, instance=instance, locked=False, group=group)

        return instance

    def deregister(self, name: str) -> Any:
        self.logger.info(f'Deregistering instance: {name}')
        self.websocket_connection.send_request_sync(messages_pb2.Method.DEREGISTER, data=[name])

    def _get_instance(self, instance_name: Union[type, str]) -> Instance:
        try:
            if isclass(instance_name):
                return self.instances[instance_name.__name__]

            return self.instances[instance_name]
        except KeyError:
            raise errors.InstanceNotFoundError(f'Instance not found: {instance_name}')

    def _available(self, instance_name: str) -> messages_pb2.Message:
        if instance_name not in self.instances:
            return utils.build_response(status=messages_pb2.Status.NOT_FOUND)

        if self.instances[instance_name].locked:
            return utils.build_response(status=messages_pb2.Status.LOCKED)

        return utils.build_response()

    def _lock(self, instance_name: str) -> messages_pb2.Message:
        instance = self._get_instance(instance_name)
        instance.locked = True

        return utils.build_response()

    def _unlock(self, instance_name: str) -> messages_pb2.Message:
        instance = self._get_instance(instance_name)
        instance.locked = False

        return utils.build_response()

    def _metadata(self, instance_name: str) -> messages_pb2.Message:
        instance = self._get_instance(instance_name)
        metadata = InstanceMetadata.from_instance(instance)

        return utils.build_response(data=asdict(metadata))

    def _get(self, instance_name: str, attribute_path: str) -> messages_pb2.Message:
        try:
            instance = self._get_instance(instance_name)
            data = get_attribute_path(instance.instance, attribute_path.split('.'))

            if callable(data):
                return utils.build_response(status=messages_pb2.Status.METHOD)

            try:
                serialized_data = pickle.dumps(data)
                return utils.build_response(serialized_data=serialized_data)
            except (pickle.PickleError, TypeError):
                return utils.build_response(status=messages_pb2.Status.PROXY)
        except Exception as err:
            return utils.build_error_response(err)

    def _set(self, instance_name: str, attribute_path: str, value: Any) -> messages_pb2.Message:
        try:
            instance = self._get_instance(instance_name)
            set_attribute_path(instance.instance, attribute_path.split('.'), value)
            return utils.build_response()
        except Exception as err:
            return utils.build_error_response(err)

    def _call(self, instance_name: str, attribute_path: str, args: Union[List[Any], Dict[str, Any]], kwargs: Dict[str, Any]) -> messages_pb2.Message:
        try:
            instance = self._get_instance(instance_name)
            method = get_attribute_path(instance.instance, attribute_path.split('.'))

            # handle special case where args is encoded as a dict with numerical string indexes
            if isinstance(args, dict):
                args = [args[str(i)] for i in range(len(args))]

            data = method(*args, **kwargs)
            return utils.build_response(data=data)
        except Exception as err:
            return utils.build_error_response(err)

    async def _run(self, *args: Any) -> Any:
        return await self.event_loop.run_in_executor(self.executor, *args)

    async def _handle_request(self, websocket_connection: websockets.WebSocketCommonProtocol, message: messages_pb2.Message) -> messages_pb2.Message:
        try:
            request = message.request
            data = utils.decode_data(message.data, message.encoding)

            if request.method == messages_pb2.Method.GET:
                return await self._run(self._get, request.instance, request.attribute)

            if request.method == messages_pb2.Method.SET:
                return await self._run(self._set, request.instance, request.attribute, data)

            if request.method == messages_pb2.Method.CALL:
                return await self._run(self._call, request.instance, request.attribute, data['args'], data['kwargs'])

            if request.method == messages_pb2.Method.AVAILABLE:
                return await self._run(self._available, request.instance)

            if request.method == messages_pb2.Method.LOCK:
                return await self._run(self._lock, request.instance)

            if request.method == messages_pb2.Method.UNLOCK:
                return await self._run(self._unlock, request.instance)

            if request.method == messages_pb2.Method.METADATA:
                return await self._run(self._metadata, request.instance)
        except KeyError:
            return utils.build_error_response(errors.InvalidMessageError(f'Invalid message'))

        return utils.build_error_response(errors.InvalidMethodError(f'Invalid method: {request.method}'))


if __name__ == '__main__':
    import time
    logging.basicConfig(level=logging.INFO)

    class TestClass:
        foo = 'bar'

        def method(self):
            return 'baz'

        def sleep(self, duration):
            time.sleep(duration)

    server = Server()
    server.register('test', TestClass())
    server.start()


from typing import Optional, Dict, Any, Callable, Awaitable, Union
import asyncio
import pickle
import json
import logging
import websockets
from concurrent import futures
from google.protobuf.message import EncodeError, DecodeError
from rpc_gateway import errors, utils, messages_pb2

logger = logging.getLogger(__name__)
RequestHandlerType = Callable[['WebsocketConnection', messages_pb2.Message], Awaitable[messages_pb2.Message]]
CloseHandlerType = Callable[['MessagePump'], Awaitable]
MAX_MESSAGE_ID = 99999
last_message_id = 0


def next_message_id() -> int:
    global last_message_id
    last_message_id += 1
    return last_message_id % MAX_MESSAGE_ID


class WebsocketConnection:
    next_id = 0

    def __init__(self,
                 connection: Optional[websockets.WebSocketCommonProtocol] = None,
                 request_handler: Optional[RequestHandlerType] = None,
                 close_handler: Optional[CloseHandlerType] = None,
                 message_queues: Optional[Dict[int, asyncio.Queue]] = None):
        self.id = WebsocketConnection.next_id
        WebsocketConnection.next_id += 1
        self.logger = logger.getChild(self.__class__.__name__)
        self.connection = connection
        self.request_handler = request_handler
        self.close_handler = close_handler
        self.event_loop = asyncio.get_event_loop()
        self.send_queue = asyncio.Queue()
        self.receive_queues = {} if message_queues is None else message_queues
        self.message_send_task: Optional[asyncio.Task] = None
        self.message_receive_task: Optional[asyncio.Task] = None
        self.running = False

    @property
    def connected(self) -> bool:
        return self.connection.open

    async def start(self, wait=True, connection: Optional[websockets.WebSocketCommonProtocol] = None):
        if connection is not None:
            self.connection = connection

        self.running = True
        self.message_send_task = asyncio.Task(self._message_receive())
        self.message_receive_task = asyncio.Task(self._message_send())

        if wait:
            await self.wait()

    async def wait(self):
        try:
            await self.message_send_task
            await self.message_receive_task
        except futures.CancelledError:
            pass

    async def stop(self):
        self.running = False
        await self.connection.close()
        await self.connection.wait_closed()
        self.message_send_task.cancel()
        self.message_receive_task.cancel()
        try:
            await self.message_receive_task
            await self.message_send_task
        except futures.CancelledError:
            pass

    def send_message_sync(self, message: messages_pb2.Message):
        utils.await_sync(self.send_message(message), self.event_loop)

    async def send_message(self, message: messages_pb2.Message):
        if self.connection is None:
            raise errors.NotConnectedError('Must be connected to send message: {message}')

        await self.send_queue.put(message)

    def send_request_sync(self, method: str, instance: Optional[str] = None, attribute: Optional[str] = None,
                          data: Any = None, encoding: messages_pb2.Encoding = messages_pb2.Encoding.PICKLE) -> messages_pb2.Message:
        return utils.await_sync(self.send_request(method, instance, attribute, data, encoding), self.event_loop)

    async def send_request(self, method: str, instance: Optional[str] = None, attribute: Optional[str] = None,
                          data: Any = None, encoding: messages_pb2.Encoding = messages_pb2.Encoding.PICKLE, raise_error: bool = True) -> messages_pb2.Message:
        encoded_data = utils.encode_data(data, encoding)
        request = messages_pb2.Request(method=method, instance=instance, attribute=attribute)
        message = messages_pb2.Message(data=encoded_data, encoding=encoding, request=request)
        return await self.request(message, raise_error=raise_error)

    def request_sync(self, request: messages_pb2.Message, raise_error=True) -> messages_pb2.Message:
        return utils.await_sync(self.request(request, raise_error))

    async def request(self, request: messages_pb2.Message, raise_error=True) -> messages_pb2.Message:
        request.id = next_message_id()
        queue = self.receive_queues[request.id] = asyncio.Queue()
        await self.send_message(request)

        try:
            response: messages_pb2.Message = await queue.get()
            self.receive_queues.pop(request.id)
        except asyncio.TimeoutError:
            self.receive_queues.pop(request.id)
            raise errors.RequestTimeoutError(f'Request timed out while waiting for response: {request}')

        # raise an error if this is an error response
        if response.response.status == messages_pb2.Status.ERROR and raise_error:
            decoded_error = utils.decode_data(response.data, response.encoding)

            if response.encoding == messages_pb2.Encoding.JSON:
                raise Exception(decoded_error)

            if response.encoding == messages_pb2.Encoding.PICKLE:
                raise decoded_error

            raise errors.InvalidEncodingError(f'Invalid response encoding: {response.encoding}')

        return response

    async def _handle_request(self, request: messages_pb2.Message):
        id = request.id
        response: messages_pb2.Message = await self.request_handler(self, request)
        response.id = id
        await self.send_message(response)

    async def _message_send(self):
        while self.running:
            message: messages_pb2.Message = await self.send_queue.get()

            try:
                message_bytes = message.SerializeToString()
            except EncodeError:
                error = errors.SerializationError(f'Cannot serialize response data')
                message = messages_pb2.Message(id=message.id, data=utils.encode_error(error, message.encoding))
                message.response.status = messages_pb2.Status.ERROR
                message_bytes = message.SerializeToString()

            self.logger.info(f'[{self.id}] > {message}')
            await self.connection.send(message_bytes)

    async def _message_receive(self):
        try:
            while self.running:
                message_bytes = await self.connection.recv()
                try:
                    message = messages_pb2.Message()
                    message.ParseFromString(message_bytes)
                except DecodeError:
                    self.logger.error(f'Invalid message: {message_bytes}')
                    continue

                self.logger.info(f'[{self.id}] < {message}')

                if message.HasField('response'):
                    if message.id not in self.receive_queues:
                        raise errors.InvalidMessageIdError(f'No request found for response ID: {message.id}')

                    queue = self.receive_queues[message.id]
                    await queue.put(message)

                elif message.HasField('request'):  # request
                    if self.request_handler is not None:
                        asyncio.create_task(self._handle_request(message))

                else:
                    self.logger.error(f'Invalid message: {message}')

        except websockets.ConnectionClosed as err:
            self.logger.info(f'[{self.id}] Connection closed')

            if isinstance(err, websockets.ConnectionClosedError):
                self.logger.error(err)

            if self.close_handler is not None:
                await self.close_handler(self)


import json
from typing import Optional, Any, Dict, List
import logging
import re
from aiohttp import web
from rpc_gateway import messages_pb2, websocket_connection, utils, errors

logger = logging.getLogger(__name__)


class MessageHandler:
    SERVER_MESSAGES = []

    def handle_forward_request(self, message: messages_pb2.Message) -> messages_pb2.Message:
        pass

    def handle_available_request(self, message: messages_pb2.Message) -> messages_pb2.Message:
        pass

    def handle_list_request(self, message: messages_pb2.Message) -> messages_pb2.Message:
        pass


class HttpGateway:
    def __init__(self, message_handler: MessageHandler, host: str, port: int, auth_key: str):
        self.logger = logger.getChild(self.__class__.__name__)
        self.host = host
        self.port = port
        self.message_handler = message_handler
        self.auth_key = auth_key
        self.app = web.Application(middlewares=[self.authenticate])
        self.runner = web.AppRunner(self.app)
        self.site: Optional[web.TCPSite] = None
        self.running = False

        self.app.add_routes([web.get('/instance/{instance}/{attribute}', self.get),
                             web.put('/instance/{instance}/{attribute}', self.set),
                             web.post('/instance/{instance}/{attribute}', self.call),
                             web.get('/available/{instance}', self.available),
                             web.get('/metadata/{instance}', self.metadata),
                             web.get('/list/{group}', self.list),
                             web.post('/', self.legacy_request)])

    async def start(self):
        self.logger.info(f'Starting HTTP server on http://{self.host}:{self.port}, auth_key={self.auth_key}')
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.host, self.port)
        await self.site.start()
        self.running = True

    async def stop(self):
        await self.site.stop()
        self.running = False

    @web.middleware
    async def authenticate(self, request: web.BaseRequest, handler) -> web.StreamResponse:
        if 'Authentication' in request.headers:
            auth_header = request.headers['Authentication']

            if auth_header.startswith('AuthKey '):
                auth_type, auth_key = auth_header.split(' ', maxsplit=2)

                if auth_key == self.auth_key:
                    return await handler(request)

        elif 'auth_key' in request.query and request.query['auth_key'] == self.auth_key:
            return await handler(request)

        return web.json_response('Unauthorized', status=401)

    async def get(self, request: web.Request) -> web.StreamResponse:
        response = await self.message_handler.handle_forward_request(utils.build_request(method=messages_pb2.Method.GET,
                                                                                         instance=request.match_info.get('instance'),
                                                                                         attribute=request.match_info.get('attribute')))

        return utils.build_http_response(response)

    async def set(self, request: web.Request) -> web.StreamResponse:
        encoded_data = utils.encode_data(request.json(), messages_pb2.Encoding.PICKLE)
        response = await self.message_handler.handle_forward_request(utils.build_request(method=messages_pb2.Method.SET,
                                                                                         instance=request.match_info.get('instance'),
                                                                                         attribute=request.match_info.get('attribute'),
                                                                                         data=encoded_data))

        return utils.build_http_response(response)

    async def call(self, request: web.Request) -> web.StreamResponse:
        data: Dict = (await request.json()) or {}
        query = {k: v for k, v in request.query.items() if k != 'AUTH_KEY'}
        kwargs = {**data.get('kwargs', {}), **query}
        args = data.get('args', [])
        encoded_data = utils.encode_data({'args': args, 'kwargs': kwargs}, messages_pb2.Encoding.PICKLE)
        response = await self.message_handler.handle_forward_request(utils.build_request(method=messages_pb2.Method.CALL,
                                                                                         instance=request.match_info.get('instance'),
                                                                                         attribute=request.match_info.get('attribute'),
                                                                                         data=encoded_data))

        return utils.build_http_response(response)

    async def available(self, request: web.Request) -> web.StreamResponse:
        response = await self.message_handler.handle_available_request(utils.build_request(method=messages_pb2.Method.AVAILABLE,
                                                                                           instance=request.match_info.get('instance')))

        return utils.build_http_response(response)

    async def metadata(self, request: web.Request) -> web.StreamResponse:
        response = await self.message_handler.handle_forward_request(utils.build_request(method=messages_pb2.Method.METADATA,
                                                                                         instance=request.match_info.get('instance')))

        return utils.build_http_response(response)

    async def list(self, request: web.Request) -> web.StreamResponse:
        data = utils.encode_data(request.match_info.get('group', None), messages_pb2.Encoding.PICKLE)
        response = await self.message_handler.handle_list_request(utils.build_request(method=messages_pb2.Method.LIST,
                                                                                      data=data))

        return utils.build_http_response(response)

    async def legacy_request(self, request: web.Request) -> web.StreamResponse:
        if request.method != 'POST':
            return web.HTTPMethodNotAllowed(request.method, ['POST'])

        message_raw = await request.read()

        try:
            self.logger.info(f'Handling request: {message_raw}')
            message = json.loads(message_raw)
            method = messages_pb2.Method.Value(message['method'].upper())
            request = messages_pb2.Message()
            request.request.method = method

            if 'instance' in message['data']:
                request.request.instance = message['data'].pop('instance')

            if 'attribute' in message['data']:
                request.request.attribute = message['data'].pop('attribute')

            request.data = utils.encode_data(message['data'], messages_pb2.Encoding.PICKLE)

            if method in self.message_handler.SERVER_MESSAGES:
                response = await self.message_handler.handle_forward_request(request)
            elif method == messages_pb2.Method.AVAILABLE:
                response = await self.message_handler.handle_available_request(request)
            elif method == messages_pb2.Method.LIST:
                response = await self.message_handler.handle_list_request(request)
            else:
                response = utils.build_error_response(errors.InvalidMethodError(f'Invalid method: {request.method}'))

        except (KeyError, json.decoder.JSONDecodeError):
            response = utils.build_error_response(errors.InvalidMessageError(f'Invalid message: {message_raw}'))

        response_dict = {'status': messages_pb2.Status.Name(response.response.status).lower()}

        if response.response.status == messages_pb2.Status.ERROR:
            response_dict['data'] = repr(utils.decode_data(response.data, response.encoding))
        elif response.data:
            response_dict['data'] = utils.decode_data(response.data, response.encoding)

        return web.json_response(response_dict)


from typing import Optional, Awaitable, Any
import asyncio
import pickle
import json
from aiohttp import web
from rpc_gateway import messages_pb2, errors


def await_sync(awaitable: Awaitable, loop: Optional[asyncio.AbstractEventLoop] = None, timeout: Optional[float] = None) -> Any:
    loop = loop or asyncio.get_event_loop()
    if loop.is_running():
        future = asyncio.run_coroutine_threadsafe(awaitable, loop or asyncio.get_event_loop())
        return future.result(timeout=timeout)
    else:
        try:
            return loop.run_until_complete(awaitable)
        except asyncio.CancelledError:
            pass


def encode_data(data: Any, encoding: messages_pb2.Encoding) -> bytes:
    if encoding == messages_pb2.Encoding.PICKLE:
        return pickle.dumps(data)

    if encoding == messages_pb2.Encoding.JSON:
        return json.dumps(data).encode()

    raise errors.InvalidEncodingError(f'Invalid message encoding: {encoding}')


def encode_error(error: Exception, encoding: messages_pb2.Encoding) -> bytes:
    if encoding == messages_pb2.Encoding.PICKLE:
        return pickle.dumps(error)

    if encoding == messages_pb2.Encoding.JSON:
        return json.dumps(str(error)).encode()

    raise errors.InvalidEncodingError(f'Invalid message encoding: {encoding}')


def decode_data(data: bytes, encoding: messages_pb2.Encoding) -> Any:
    if len(data) == 0:
        return None

    if encoding == messages_pb2.Encoding.PICKLE:
        return pickle.loads(data)

    if encoding == messages_pb2.Encoding.JSON:
        return json.loads(data.decode())

    raise errors.InvalidEncodingError(f'Invalid message encoding: {encoding}')


def build_response(data: Optional[Any] = None, serialized_data: Optional[bytes] = None, status: Optional['messages_pb2.Status.V'] = messages_pb2.Status.SUCCESS,
                   id: Optional[int] = None, encoding: Optional['messages_pb2.Encoding.V'] = None):
    message = messages_pb2.Message(encoding=encoding)
    message.response.status = status

    if data is not None:
        message.data = pickle.dumps(data)

    if serialized_data is not None:
        message.data = serialized_data

    if id is not None:
        message.id = id

    return message


def build_error_response(error: Optional[Exception]) -> messages_pb2.Message:
    return build_response(status=messages_pb2.Status.ERROR, data=error)


def build_request(method: str, instance: Optional[str] = None, attribute: Optional[str] = None, id: Optional[int] = None,
                  data: Any = None, encoding: Optional['messages_pb2.Encoding.V'] = None) -> messages_pb2.Message:
    request = messages_pb2.Request(method=method, instance=instance, attribute=attribute)
    return messages_pb2.Message(id=id, data=data, encoding=encoding, request=request)


def build_http_response(response: messages_pb2.Message) -> web.Response:
    data = decode_data(response.data, response.encoding)
    return web.json_response(data, status=http_status(response.response.status), headers={'RPC-Status': str(response.response.status)})


def http_status(response_status: 'messages_pb2.Status.V') -> int:
    if response_status == messages_pb2.Status.NOT_FOUND:
        return 404

    if response_status == messages_pb2.Status.LOCKED:
        return 423

    if response_status == messages_pb2.Status.ERROR:
        return 503

    return 200


from typing import Optional, List, Dict, Any
import logging
import time
import asyncio
import websockets
import json
import urllib
import threading
import re
from concurrent.futures._base import CancelledError
from dataclasses import asdict
from aiohttp import web
from rpc_gateway import errors, messages_pb2, utils, http_gateway
from rpc_gateway.websocket_connection import WebsocketConnection, next_message_id
from rpc_gateway.utils import await_sync

logger = logging.getLogger(__name__)


class Gateway(http_gateway.MessageHandler):
    SERVER_MESSAGES = (messages_pb2.Method.GET, messages_pb2.Method.SET, messages_pb2.Method.CALL, messages_pb2.Method.LOCK,
                       messages_pb2.Method.UNLOCK, messages_pb2.Method.METADATA)

    def __init__(self, host: str = 'localhost', port: int = 8888, http_port: int = 8887, auth_key: str = 'DEFAULT_KEY', event_loop: Optional[asyncio.AbstractEventLoop] = None):
        self.host = host
        self.port = port
        self.auth_key = auth_key
        self.logger = logger.getChild(self.__class__.__name__)
        self.websocket_connections: List[WebsocketConnection] = []
        self.websocket: Optional[websockets.WebSocketServer] = None
        self.http_gateway = http_gateway.HttpGateway(self, host, http_port, auth_key)
        self.event_loop = event_loop or asyncio.get_event_loop()
        self.instances: Dict[str, WebsocketConnection] = {}
        self.instance_groups: Dict[str, List[str]] = {}
        self.websocket_instances: Dict[WebsocketConnection, List[str]] = {}
        self.check_connections_task: Optional[asyncio.Task] = None
        self.running = False

    def start(self, wait = True):
        self.event_loop.run_until_complete(self._start(wait=False))

        if wait:
            await_sync(self._wait())

    async def _start(self, wait = True):
        self.logger.info(f'Starting on ws://{self.host}:{self.port}')
        self.websocket = await websockets.serve(self.on_connection, self.host, self.port)
        await self.http_gateway.start()
        self.running = True
        self.check_connections_task = self.check_connections()

        if wait:
            await self._wait()

    def wait(self):
        await_sync(self._wait(), self.event_loop)

    async def _wait(self):
        if self.websocket is not None:
            await self.websocket.wait_closed()

        self.logger.info(f'Done')

    def stop(self):
        await_sync(self._stop(), self.event_loop)

    async def _stop(self):
        self.running = False
        await self.http_gateway.stop()
        try:
            await asyncio.gather(*[websocket_connection.stop() for websocket_connection in self.websocket_connections])
        except asyncio.exceptions.CancelledError:
            pass
        self.websocket.close()
        await self.check_connections_task

    async def on_connection(self, connection: websockets.WebSocketServerProtocol, path: str):
        self.logger.info(f'New connection from {connection.remote_address} path: {path}')
        connection_params = urllib.parse.parse_qs(path[1:]) if path != '' else {}

        if 'auth_key' not in connection_params or connection_params['auth_key'][0] != self.auth_key:
            await connection.close(1008, 'Invalid auth_key')  # 1008 is the "Policy Violation" websocket status code
            return

        websocket_connection = WebsocketConnection(connection, request_handler=self.handle_request, close_handler=self.handle_close)
        self.websocket_connections.append(websocket_connection)
        self.websocket_instances[websocket_connection] = []

        await websocket_connection.start()

    async def _unlock_instance(self, instance_name: str) -> messages_pb2.Message:
        server = self.instances[instance_name]
        return await server.send_request(messages_pb2.Method.UNLOCK, instance=instance_name, raise_error=False)

    def is_registered(self, group: str, instance: str) -> bool:
        if group not in self.instance_groups:
            return False

        return instance in self.instance_groups[group]

    async def check_connections(self):
        while self.running:
            for websocket_connection in self.websocket_connections:
                if not websocket_connection.connected:
                    await self.handle_close(websocket_connection)

            time.sleep(1)

    #
    # Request Handlers
    #

    async def handle_forward_request(self, request: messages_pb2.Message) -> messages_pb2.Message:
        self.logger.info(f'Forwarding request to server: {request}')
        instance_name = request.request.instance

        if instance_name not in self.instances:
            return utils.build_error_response(errors.InstanceNotFoundError(f'Instance not found: {instance_name}'))

        server = self.instances[instance_name]

        response = await server.request(request, raise_error=False)
        self.logger.info(f'Forwarding response to client: {response}')

        return response

    async def handle_available_request(self, request: messages_pb2.Message) -> messages_pb2.Message:
        if request.request.instance not in self.instances:
            return utils.build_response(status=messages_pb2.Status.NOT_FOUND)

        server = self.instances[request.request.instance]
        return await server.request(request)

    async def handle_list_request(self, request: messages_pb2.Message) -> messages_pb2.Message:
        group = utils.decode_data(request.data, request.encoding)

        if not group:
            instances = self.instance_groups
        else:
            if group not in self.instance_groups:
                return utils.build_error_response(errors.GroupNotFoundError(f'Group not found: {group}'))

            instances = {group: self.instance_groups[group]}

        return utils.build_response(instances)

    async def handle_register_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
        instances = utils.decode_data(request.data, request.encoding)

        self.logger.info(f'Registering instances: {instances}')

        # check to see if this instance
        for instance_name, instance_group in instances:
            if instance_name is None:
                return utils.build_error_response(errors.InvalidMessageError(f'Cannot register instance with name: {instance_name}'))

            if instance_group is None:
                return utils.build_error_response(errors.InvalidMessageError(f'Cannot register instance with group: {instance_group}'))

            if self.is_registered(instance_group, instance_name):
                return utils.build_error_response(errors.InstanceAlreadyRegisteredError(f'Instance "{instance_name}" already registered with "{instance_group}" group'))

        for instance_name, instance_group in instances:
            if instance_group not in self.instance_groups:
                self.instance_groups[instance_group] = []

            self.instance_groups[instance_group].append(instance_name)
            self.instances[instance_name] = websocket_connection
            self.websocket_instances[websocket_connection].append(instance_name)

        return utils.build_response()

    async def handle_deregister_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
        instances = utils.decode_data(request.data, request.encoding)
        self.logger.info(f'Deregistering instances: {instances}')

        # find instance to deregister
        for instance_name in instances:
            if instance_name not in self.instances:
                return utils.build_error_response(errors.InstanceNotFoundError(f'Instance "{instance_name}" not found'))

            instance_connection = self.instances[instance_name]
            self.instances.pop(instance_name)
            self.websocket_instances[instance_connection].remove(instance_name)

            for group, group_instances in self.instance_groups.items():
                if instance_name in group_instances:
                    group_instances.remove(instance_name)

        return utils.build_response()

    # this is called by the WebsocketConnection MessagePump when a new request is received
    async def handle_request(self, websocket_connection: WebsocketConnection, message: messages_pb2.Message) -> messages_pb2.Message:
        method = message.request.method

        if method in self.SERVER_MESSAGES:
            return await self.handle_forward_request(message)

        if method == messages_pb2.Method.AVAILABLE:
            return await self.handle_available_request(message)

        if method == messages_pb2.Method.REGISTER:
            return await self.handle_register_request(websocket_connection, message)

        if method == messages_pb2.Method.DEREGISTER:
            return await self.handle_deregister_request(websocket_connection, message)

        if method == messages_pb2.Method.LIST:
            return await self.handle_list_request(message)

        return utils.build_error_response(errors.InvalidMethodError(f'Invalid method: {method}'))

    async def handle_close(self, websocket_connection: WebsocketConnection):
        # check if this is a server connection
        if websocket_connection in self.websocket_instances:
            instances = self.websocket_instances[websocket_connection]

            # remove registered instances
            for instance_name in instances:
                if instance_name in self.instances:
                    self.instances.pop(instance_name)

                groups_to_remove = []

                for group, group_instances in self.instance_groups.items():
                    if instance_name in group_instances:
                        group_instances.remove(instance_name)

                    if len(group_instances) == 0:
                        groups_to_remove.append(group)

                for group in groups_to_remove:
                    if group in self.instance_groups:
                        self.instance_groups.pop(group)

            self.websocket_instances.pop(websocket_connection)

        # send an error response for any in-progress requests
        for request_id, response_queue in websocket_connection.receive_queues.items():
            await response_queue.put(utils.build_error_response(errors.ServerConnectionLostError(f'Server connection lost')))

        self.websocket_connections.remove(websocket_connection)

        self.logger.info(f'Connection from {websocket_connection.connection.remote_address} closed')


class GatewayClient:
    def __init__(self, gateway_url: str = 'ws://localhost:8888', auth_key: str = 'DEFAULT_KEY'):
        self.logger = logger.getChild(self.__class__.__name__)
        self.gateway_url = gateway_url[:-1] if gateway_url.endswith('/') else gateway_url
        self.gateway_auth_key = auth_key
        self.websocket_connection = WebsocketConnection(request_handler=self._log_and_handle_request, close_handler=self._handle_close)
        self.event_loop = asyncio.get_event_loop()
        self.connect_retry_timeout = 2.0
        self.start_thread = threading.Thread(name='RPC gateway client', target=self._start_sync, daemon=True)

    @property
    def connected(self) -> bool:
        return self.websocket_connection.connection is not None

    async def _connect(self):
        while True:
            try:
                self.logger.info(f'Connecting to {self.gateway_url}')
                self.connection = await websockets.connect(self.gateway_url + '/socket&auth_key=' + self.gateway_auth_key)
                return
            except OSError:
                self.logger.warning(f'Error connecting to {self.gateway_url}, retrying in {self.connect_retry_timeout} seconds')
                time.sleep(self.connect_retry_timeout)

    def start(self, wait=True):
        self.start_thread.start()
        if wait:
            self.start_thread.join()

    def _start_sync(self):
        await_sync(self._start(), self.event_loop)

    async def _start(self):
        await self._connect()
        await self.websocket_connection.start(wait=False, connection=self.connection)
        await self._on_start()
        await self._wait()

    def wait(self):
        self.start_thread.join()

    async def _wait(self):
        await self.websocket_connection.wait()

    async def _on_start(self):
        pass

    async def _on_stop(self):
        pass

    def stop(self):
        try:
            await_sync(self._stop(), self.event_loop)
        except CancelledError:
            pass

        self.start_thread.join()

    async def _stop(self):
        await self._on_stop()
        await self.websocket_connection.stop()

    async def _log_and_handle_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
        self.logger.debug(f'Request from {websocket_connection}: {request}')
        response = await self._handle_request(websocket_connection, request)
        self.logger.debug(f'Response to {websocket_connection}: {response}')

        return response

    async def _handle_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
        pass

    async def _handle_close(self, websocket_connection: WebsocketConnection):
        await self._connect()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    gateway = Gateway()
    gateway.start()


from typing import Any, Iterable, Optional, List, Dict, Callable
import time
import logging
import asyncio
from rpc_gateway import errors, gateway, messages_pb2, utils


class Client(gateway.GatewayClient):
    @staticmethod
    def join_attribute_paths(*paths):
        return '.'.join([path for path in paths if path != ''])

    def connect(self, timeout: float = 10.0):
        if not self.connected:
            self.start(wait=False)

            start_time = time.time()
            while not self.connected:
                if time.time() - start_time > timeout:
                    raise errors.ConnectionError(f'Timeout while waiting for client to connect')

                time.sleep(0.01)

    def get_instance(self, instance_name, read_only: bool = False) -> Any:
        if not read_only:
            if self.instance_locked(instance_name):
                raise errors.InstanceLockedError(f'Instance locked: {instance_name}')

            self.lock(instance_name)

        return self.proxy_attribute(instance_name, '', read_only=read_only)

    def proxy_method(self, instance_name: str, method_path: str, read_only: bool = False) -> Callable:
        def _proxy_method(*args, **kwargs):
            if read_only:
                raise errors.InstanceReadOnly(f'Cannot call method {method_path} on read-only instance')

            return self.call(instance_name, method_path, args, kwargs)

        return _proxy_method

    def proxy_attribute(self, instance_name: str, attribute_path: str, read_only: bool = False) -> object:
        client = self

        class _Proxy:
            def __dir__(self) -> Iterable[str]:
                return client.call(instance_name, client.join_attribute_paths(attribute_path, '__dir__'))

            def __getattr__(self, key):
                return client.get(instance_name, client.join_attribute_paths(attribute_path, key))

            def __setattr__(self, key, value):
                if read_only:
                    raise errors.InstanceReadOnly(f'Cannot set attribute {key} on read-only instance')

                return client.set(instance_name, client.join_attribute_paths(attribute_path, key), value)

            def __repr__(self):
                return client.call(instance_name, client.join_attribute_paths(attribute_path, '__repr__'))

            def __str__(self):
                return client.call(instance_name, client.join_attribute_paths(attribute_path, '__str__'))

        return _Proxy()

    def instance_exists(self, instance_name: str) -> bool:
        self.connect()

        response = self.websocket_connection.send_request_sync(messages_pb2.Method.AVAILABLE, instance=instance_name)

        return response.response.status != messages_pb2.Status.NOT_FOUND

    def instance_available(self, instance_name: str) -> bool:
        self.connect()

        response = self.websocket_connection.send_request_sync(messages_pb2.Method.AVAILABLE, instance=instance_name)
        return response.response.status == messages_pb2.Status.SUCCESS

    def instance_locked(self, instance_name: str) -> bool:
        self.connect()

        response = self.websocket_connection.send_request_sync(messages_pb2.Method.AVAILABLE, instance=instance_name)

        if response.response.status == messages_pb2.Status.NOT_FOUND:
            raise errors.InstanceNotFoundError(f'Instance not found: {instance_name}')

        return response.response.status == messages_pb2.Status.LOCKED

    def lock(self, instance_name: str):
        self.connect()

        self.websocket_connection.send_request_sync(messages_pb2.Method.LOCK, instance=instance_name)

    def unlock(self, instance_name: str):
        self.connect()

        self.websocket_connection.send_request_sync(messages_pb2.Method.UNLOCK, instance=instance_name)

    def metadata(self, instance_name: str):
        self.connect()

        response = self.websocket_connection.send_request_sync(messages_pb2.Method.METADATA, instance=instance_name)
        return utils.decode_data(response.data, response.encoding)

    def list(self, group: Optional[str]=None) -> Dict[str, List[str]]:
        self.connect()

        response = self.websocket_connection.send_request_sync(messages_pb2.Method.LIST, data=group)
        return utils.decode_data(response.data, response.encoding)

    def call(self, instance_name: str, method_path: str, args: Optional[Iterable[Any]] = None, kwargs: Optional[Dict[str, Any]] = None) -> Any:
        self.connect()

        data = {'args': args or [], 'kwargs': kwargs or {}}
        response = self.websocket_connection.send_request_sync(messages_pb2.Method.CALL, instance=instance_name, attribute=method_path, data=data)

        return utils.decode_data(response.data, response.encoding)

    def get(self, instance_name: str, attribute_path: str, read_only: bool = False) -> Any:
        self.connect()

        response = self.websocket_connection.send_request_sync(messages_pb2.Method.GET, instance=instance_name, attribute=attribute_path)

        if response.response.status == messages_pb2.Status.METHOD:
            return self.proxy_method(instance_name, attribute_path, read_only=read_only)

        if response.response.status == messages_pb2.Status.PROXY:
            return self.proxy_attribute(instance_name, attribute_path)

        return utils.decode_data(response.data, response.encoding)

    def set(self, instance_name: str, attribute_path: str, value: Any) -> Any:
        self.connect()

        self.websocket_connection.send_request_sync(messages_pb2.Method.SET, instance=instance_name, attribute=attribute_path, data=value)

        return value


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    class TestClass:
        foo = 'bar'

        def method(self):
            return 'baz'

    client = Client()
    test: TestClass = client.get_instance('test')
    print(test.foo)
    # print(test.method())
    # print(test)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0emessages.proto\"\x87\x01\n\x07Message\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x1b\n\x08\x65ncoding\x18\x02 \x01(\x0e\x32\t.Encoding\x12\x1b\n\x07request\x18\x05 \x01(\x0b\x32\x08.RequestH\x00\x12\x1d\n\x08response\x18\x06 \x01(\x0b\x32\t.ResponseH\x00\x12\x0c\n\x04\x64\x61ta\x18\t \x01(\x0c\x42\t\n\x07message\"G\n\x07Request\x12\x17\n\x06method\x18\x01 \x01(\x0e\x32\x07.Method\x12\x10\n\x08instance\x18\x02 \x01(\t\x12\x11\n\tattribute\x18\x03 \x01(\t\"#\n\x08Response\x12\x17\n\x06status\x18\x01 \x01(\x0e\x32\x07.Status* \n\x08\x45ncoding\x12\n\n\x06PICKLE\x10\x00\x12\x08\n\x04JSON\x10\x01*R\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x12\n\n\x06METHOD\x10\x02\x12\t\n\x05PROXY\x10\x03\x12\r\n\tNOT_FOUND\x10\x04\x12\n\n\x06LOCKED\x10\x05*\x7f\n\x06Method\x12\x07\n\x03GET\x10\x00\x12\x07\n\x03SET\x10\x01\x12\x08\n\x04\x43\x41LL\x10\x02\x12\x0c\n\x08REGISTER\x10\x03\x12\x0e\n\nDEREGISTER\x10\x04\x12\x08\n\x04LOCK\x10\x05\x12\n\n\x06UNLOCK\x10\x06\x12\r\n\tAVAILABLE\x10\x07\x12\x0c\n\x08METADATA\x10\x08\x12\x08\n\x04LIST\x10\tb\x06proto3'
)

_ENCODING = _descriptor.EnumDescriptor(
  name='Encoding',
  full_name='Encoding',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PICKLE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JSON', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=266,
  serialized_end=298,
)
_sym_db.RegisterEnumDescriptor(_ENCODING)

Encoding = enum_type_wrapper.EnumTypeWrapper(_ENCODING)
_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='METHOD', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROXY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NOT_FOUND', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOCKED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=300,
  serialized_end=382,
)
_sym_db.RegisterEnumDescriptor(_STATUS)

Status = enum_type_wrapper.EnumTypeWrapper(_STATUS)
_METHOD = _descriptor.EnumDescriptor(
  name='Method',
  full_name='Method',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GET', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SET', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CALL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REGISTER', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DEREGISTER', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOCK', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNLOCK', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='AVAILABLE', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='METADATA', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LIST', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=384,
  serialized_end=511,
)
_sym_db.RegisterEnumDescriptor(_METHOD)

Method = enum_type_wrapper.EnumTypeWrapper(_METHOD)
PICKLE = 0
JSON = 1
SUCCESS = 0
ERROR = 1
METHOD = 2
PROXY = 3
NOT_FOUND = 4
LOCKED = 5
GET = 0
SET = 1
CALL = 2
REGISTER = 3
DEREGISTER = 4
LOCK = 5
UNLOCK = 6
AVAILABLE = 7
METADATA = 8
LIST = 9



_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Message.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='encoding', full_name='Message.encoding', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request', full_name='Message.request', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response', full_name='Message.response', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='Message.data', index=4,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='message', full_name='Message.message',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=19,
  serialized_end=154,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='Request.method', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='instance', full_name='Request.instance', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute', full_name='Request.attribute', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=227,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='Response.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=229,
  serialized_end=264,
)

_MESSAGE.fields_by_name['encoding'].enum_type = _ENCODING
_MESSAGE.fields_by_name['request'].message_type = _REQUEST
_MESSAGE.fields_by_name['response'].message_type = _RESPONSE
_MESSAGE.oneofs_by_name['message'].fields.append(
  _MESSAGE.fields_by_name['request'])
_MESSAGE.fields_by_name['request'].containing_oneof = _MESSAGE.oneofs_by_name['message']
_MESSAGE.oneofs_by_name['message'].fields.append(
  _MESSAGE.fields_by_name['response'])
_MESSAGE.fields_by_name['response'].containing_oneof = _MESSAGE.oneofs_by_name['message']
_REQUEST.fields_by_name['method'].enum_type = _METHOD
_RESPONSE.fields_by_name['status'].enum_type = _STATUS
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
DESCRIPTOR.enum_types_by_name['Encoding'] = _ENCODING
DESCRIPTOR.enum_types_by_name['Status'] = _STATUS
DESCRIPTOR.enum_types_by_name['Method'] = _METHOD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:Message)
  })
_sym_db.RegisterMessage(Message)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:Request)
  })
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
  'DESCRIPTOR' : _RESPONSE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:Response)
  })
_sym_db.RegisterMessage(Response)


# @@protoc_insertion_point(module_scope)
